import numpy as np
from cpsplines.psplines.bspline_basis import BsplineBasis


class PenaltyMatrix:

    """
    Constructs the penalty matrix of the P-splines (Eilers, P. H., Marx, B. D.
    (1996)) given a B-spline basis. Standard P-splines use a penalty that is
    based on repeated differences (`variation`='diff').

    Parameters
    ----------
    bspline : BsplineBasis
        The B-spline basis object.
    variation : str, default = 'diff'
        The penalty variation. A summary of different penalties can be found in
        Eilers, P. H., Marx, B. D., & Durbán, M. (2015).

    References
    ----------
    - Eilers, P. H., & Marx, B. D. (1996). Flexible smoothing with B-splines and
      penalties. Statistical science, 11(2), 89-121.
    - Eilers, P. H., Marx, B. D., & Durbán, M. (2015). Twenty years of
      P-splines. SORT: statistics and operations research transactions, 39(2),
      149-186.
    """

    def __init__(self, bspline: BsplineBasis, variation: str = "diff"):
        self.bspline = bspline
        self.variation = variation

    def _get_diff_matrix(self, ord_d: int = 2) -> np.ndarray:

        """
        Generate the penalty matrix based on finite differences of the
        coefficients of adjacent B-splines

        Parameters
        ----------
        ord_d : int
            The order of the penalty. Must be a non-negative integer. By
            default, 2.

        Returns
        -------
        np.ndarray of shape (`bspline.n_int` + `bspline.int_forw` +
        `bspline.int_back` + `bspline.deg`, `bspline.n_int` + `bspline.int_forw`
        + `bspline.int_back` + `bspline.deg`)
            The penalty matrix of 'diff' variation

        Raises
        ------
        ValueError
            If penalty order is less than B-spline basis degree.
        ValueError
            If `ord_d` is not a non-negative integer.
        """

        if self.bspline.deg <= ord_d:
            raise ValueError(
                "The penalty order must be less than the B-spline basis degree."
            )
        if ord_d < 0:
            raise ValueError("The penalty order must be a non-negative integer.")

        dim = (
            self.bspline.n_int
            + self.bspline.int_forw
            + self.bspline.int_back
            + self.bspline.deg
            + ord_d
        )
        # Generate an identity matrix of order `bspline.n_int` +
        # `bspline.int_forw` +  `bspline.int_back` + `bspline.deg` and generate
        # its difference matrix. Then, remove the first and last `ord_d` rows
        D = np.diff(np.eye(dim, dtype=np.int32), n=ord_d)[ord_d:-ord_d, :]
        return D.T @ D

    def get_penalty_matrix(self, **kwargs) -> np.ndarray:

        """Generates the penalty matrix associated with the B-spline basis
        `bspline` and the variation `variation`. If the chosen variation needs
        some parameters to be defined, they must be included on the kwargs.

        Returns
        -------
        np.ndarray of shape (`bspline.n_int` + `bspline.int_forw` +
        `bspline.int_back` + `bspline.deg`, `bspline.n_int` + `bspline.int_forw`
        + `bspline.int_back` + `bspline.deg`)
            The penalty matrix.

        Raises
        ------
        ValueError
            Variation type must be on ("diff", )
        """

        if self.variation not in ("diff",):
            raise ValueError("Penalty matrix type not valid.")
        if self.variation == "diff":
            return self._get_diff_matrix(**kwargs)



import pandas as pd
from typing import Iterable, Union


def print_grid_search_results(
    x_val: Iterable[Iterable[Union[int, float]]],
    obj_val: Iterable[Union[int, float]],
    top_n: int,
) -> None:
    """
    Prints the best `top_n` combinations of values that minimizes the
    objective function value.

    Parameters
    ----------
    x_val : Iterable[Iterable[Union[int, float]]]
        An iterable containing the arrays used to evaluate the objective
        function.
    obj_val : Iterable[Union[int, float]]
        The objective function values, with the same order as `x_val`
    top_n : int
        The number of top results to be shown.
    """
    df = pd.DataFrame(x_val, columns=[f"sp{i+1}" for i, _ in enumerate(x_val[0])])
    df.loc[:, "Objective"] = obj_val
    df = df.sort_values(by="Objective")
    print(f"Top {top_n} combinations minimizing the GCV criterion")
    print(*df.columns, sep="\t\t")
    for i in range(min(top_n, df.shape[0])):
        print(*df.iloc[i, :], sep="\t\t")
    return None



from functools import reduce
from typing import Iterable

import mosek.fusion
import numpy as np
from cpsplines.utils.fast_kron import fast_kronecker_product


def matrix_by_tensor_product_mosek(
    matrices: Iterable[np.ndarray], mosek_var: mosek.fusion.Variable
) -> mosek.fusion.Expr:

    """
    Performs the product of a multidimensional MOSEK decision variable with
    shape k_1 x ··· x k_N by a set of matrices A_i with shapes s_i x k_i along
    each mode. The decision variable is converted to a matrix along the 0-th
    mode, denoted by T_0, and the product A_0 @ T_0 (A_1 kron ··· kron A_N).T is
    computed. Finally, the resulting multidimensional array is reshaped so the
    final shape is (s_0, s_1, ..., s_N).

    Parameters
    ----------
    matrices : Iterable[np.ndarray]
        The matrices A_i with shapes s_i x k_i to be multiplied by the
        multidimensional decision variable.
    mosek_var : mosek.fusion.Variable
        The k_1 x ··· x k_N MOSEK decision variable.

    Returns
    -------
    mosek.fusion.Expr with shape (s_0, s_1, ..., s_N)
        The output of the product.

    Raises
    ------
    ValueError
        If any matrix in `matrices` is one-dimensional.

    """

    if any(x.ndim != 2 for x in matrices):
        raise ValueError("Matrices must be two-dimensional. No vectors allowed.")

    init_shape = mosek_var.getShape()
    # The decision variable is unfolding along the 0-th mode and the product
    # A_0 @ T_0 is performed
    out = mosek.fusion.Expr.mul(
        matrices[0], mosek_var.reshape([init_shape[0], np.prod(init_shape[1:])])
    )
    # If there is more than one input matrix, the previous out is multiplied by
    # (A_1 kron ··· kron A_N).T, where kron denotes the Kronecker product
    if matrices[1:]:
        right_mul = reduce(fast_kronecker_product, matrices[1:]).T
        out = mosek.fusion.Expr.mul(out, right_mul)
    # The output is folded along 0-th mode
    final_shape = [mat.shape[0] for mat in matrices]
    return mosek.fusion.Expr.reshape(out, final_shape)



from functools import reduce

import numpy as np
import pytest
from cpsplines.utils.fast_kron import (
    fast_kronecker_product,
    kronecker_matrix_by_identity,
)
from cpsplines.utils.timer import timer
from scipy.linalg import block_diag


# Take the Kronecker product of a list of random matrices with different shapes
# with a fast method and the NumPy method
@pytest.mark.parametrize(
    "dim",
    [
        ((4, 7), (5, 6), (8, 3), (1, 9), (10, 2)),
    ],
)
def test_fast_kronecker(dim):
    matrices = [np.random.rand(*d) for d in dim]

    with timer(tag="Using np.kron"):
        exp_out = reduce(np.kron, matrices)

    with timer(tag="Using broadcasting"):
        out = reduce(fast_kronecker_product, matrices)

    np.testing.assert_allclose(exp_out, out)


# Take the Kronecker product of a random matrix and an identity matrix
# with a fast method, the fast Kronecker product and the NumPy method
@pytest.mark.parametrize(
    "A, n",
    [
        (np.random.rand(13, 17), 19),
        (np.random.rand(53, 59), 62),
    ],
)
def test_mat_kron_identity(A, n):

    with timer(tag="Using fast kronecker product by identity"):
        out = kronecker_matrix_by_identity(A, n)

    with timer(tag="Using np.kron"):
        exp_out = np.kron(A, np.eye(n))

    with timer(tag="Using broadcasting"):
        exp_out2 = fast_kronecker_product(A=A, B=np.eye(n))

    np.testing.assert_allclose(exp_out, out)
    np.testing.assert_allclose(exp_out2, out)


# Take the Kronecker product of an identity matrix and a random matrix
# with a fast method, the fast Kronecker product and the NumPy method
@pytest.mark.parametrize(
    "n, A",
    [
        (19, np.random.rand(13, 17)),
        (62, np.random.rand(53, 59)),
    ],
)
def test_identity_kron_mat(n, A):

    with timer(tag="Using fast kronecker product identity times matrix"):
        out = block_diag(*[A] * n)

    with timer(tag="Using np.kron"):
        exp_out = np.kron(np.eye(n), A)

    with timer(tag="Using broadcasting"):
        exp_out2 = fast_kronecker_product(A=np.eye(n), B=A)

    np.testing.assert_allclose(exp_out, out)
    np.testing.assert_allclose(exp_out2, out)



from typing import Iterable, Union

import numpy as np
import pytest
from cpsplines.utils.fast_kron import penalization_term
from cpsplines.utils.timer import timer


def penalty_brute_force(
    D_mul: Iterable[np.ndarray], sp: Iterable[Union[int, float]]
) -> np.ndarray:

    """
    Computes the penalization term using the exact definition including all
    the matrices involved. It only supports three or less dimensions.

    Parameters
    ----------
    D_mul : Iterable[np.ndarray]
        The univariate penalty matrices. Must have length less or equal than 3.
    sp : Iterable[Union[int, float]]
        The smoothing parameter vector. Must have length less or equal than 3.

    Returns
    -------
    np.ndarray
        The penalization term.

    Raises
    ------
    ValueError
        If the number of input smoothing paramters and penalty matrices differ.
    ValueError
        If the number of input penalty matrices is greater than 3.
    """

    if len(sp) != len(D_mul):
        raise ValueError(
            "Number of smoothing parameters must be equal to number of D elements."
        )
    if len(D_mul) > 3:
        raise ValueError("Not implemented for more than three dimensions.")

    dim = [P.shape[0] for P in D_mul]
    if len(dim) == 1:
        P = np.multiply(sp[0], D_mul[0])
    elif len(dim) == 2:
        P = np.kron(np.multiply(sp[0], D_mul[0]), np.eye(dim[1])) + np.kron(
            np.eye(dim[0]), np.multiply(sp[1], D_mul[1])
        )
    else:
        P = (
            np.kron(
                np.multiply(sp[0], D_mul[0]),
                np.eye(dim[1] * dim[2]),
            )
            + np.kron(
                np.eye(dim[0]),
                np.kron(np.multiply(sp[1], D_mul[1]), np.eye(dim[2])),
            )
            + np.kron(np.eye(dim[0] * dim[1]), np.multiply(sp[2], D_mul[2]))
        )
    return P


# Test the computation of the penalization term in the objective function for a
# variety of number of dimensions
@pytest.mark.parametrize(
    "dim, sp",
    [
        ([41], [17]),
        ([37, 5], [19, 29]),
        ([23, 11, 17], [31, 67, 93]),
    ],
)
def test_penalty_matrix(dim, sp):
    matrices = [np.random.rand(d, d) for d in dim]

    with timer(tag="Using brute force computation"):
        penalty_out = penalty_brute_force(D_mul=matrices, sp=sp)

    with timer(tag="Using fast kronecker products"):
        penalty = np.add.reduce(
            [
                np.multiply(s, A)
                for s, A in zip(sp, penalization_term(matrices=matrices))
            ]
        )
    np.testing.assert_allclose(penalty_out, penalty)



from functools import reduce
from typing import Iterable, Tuple, Union

import numpy as np
from cpsplines.utils.fast_kron import (
    fast_kronecker_product,
    matrix_by_tensor_product,
    penalization_term,
)


def gcv_mat(
    B_mul: Iterable[np.ndarray], D_mul: Iterable[np.ndarray]
) -> Tuple[np.ndarray]:

    """
    Generate the matrices used to compute the hat matrix on the Generalized
    Cross Validation.

    Parameters
    ----------
    B_mul : Iterable[np.ndarray]
        An iterable containing the products B_i.T @ B_i, where B_i is the design
        matrix of the B-spline basis along the i-th axis.
    D_mul : Iterable[np.ndarray]
        An iterable containing the penalty matrix P_i, where P_i is the penalty
        matrix along the i-th axis (without multiplying the smoogthing
        parameters).

    Returns
    -------
    Tuple[np.ndarray]
        The matrices used in the GCV computation. The first is the B-spline
        bases related matrix and the rest are the penalty matrices.
    """
    # The order must be reversed since vec(Y) = Bvec(Theta), where the
    # vectorization is performed stacking the columns, so B is the Kronecker
    # product of B_N kron ··· kron B_1
    bases_term = reduce(fast_kronecker_product, B_mul[::-1])
    penalty_term = penalization_term(matrices=D_mul[::-1])
    return tuple([bases_term] + penalty_term)


def quadratic_term(
    sp: Iterable[Union[int, float]], Q_matrices: Iterable[np.ndarray]
) -> np.ndarray:

    """
    Given a smoothing parameter vector and the set of matrices used in the
    computation of the hat matrix, computes the matrix to be inverted in this
    computation.

    Parameters
    ----------
    sp : Iterable[Union[int, float]]
        The smoothing paramater vector.
    Q_matrices : Iterable[np.ndarray]
        The array of matrices of the matrices to be inverted. The first is
        related to the B-spline bases and the rest to the penalty terms along
        each dimension.

    Returns
    -------
    np.ndarray
        The matrix to be inverted on the computation of the Generalized Cross
        Validation.

    Raises
    ------
    ValueError
        If matrices dimensions are different.
    """

    if len(set([A.shape for A in Q_matrices])) > 1:
        raise ValueError("Matrices dimensions must agree.")
    penalty_term = np.zeros(Q_matrices[0].shape)
    reversed_sp = sp[::-1]
    for i, s in enumerate(reversed_sp):
        penalty_term += np.multiply(s, Q_matrices[i + 1])
    return Q_matrices[0] + penalty_term


def explicit_y_hat(
    Q: np.ndarray, B_weighted: Iterable[np.ndarray], y: np.ndarray
) -> np.ndarray:

    """
    Computes the fitted values for the response variable using the explicit
    formula on the unconstrained framework.

    Parameters
    ----------
    Q : np.ndarray
        The resulting matrix from `quadratic_term`.
    B_weighted : Iterable[np.ndarray]
        The weighted design matrices from B-spline basis.
    y : np.ndarray
        The response variable sample.

    Returns
    -------
    np.ndarray
        The fitted values for the response variable.
    """
    y_contribution = matrix_by_tensor_product([B.T for B in B_weighted], y).flatten("F")
    theta = np.reshape(
        np.linalg.solve(Q, y_contribution),
        tuple([mat.shape[1] for mat in B_weighted]),
        order="F",
    )
    return matrix_by_tensor_product([mat for mat in B_weighted], theta)


def GCV(
    sp: Iterable[Union[int, float]],
    B_weighted: Iterable[np.ndarray],
    Q_matrices: Iterable[np.ndarray],
    y: np.ndarray,
) -> float:

    """Computes the Generalized Cross Validation (Golub et al., 1979).

    Parameters
    ----------
    sp : Iterable[Union[int, float]]
        The smoothing paramater vector.
    B_weighted : Iterable[np.ndarray]
        The weighted design matrices from B-spline basis.
    Q_matrices : Iterable[np.ndarray]
        The array of matrices of the matrices to be inverted. The first is
        related to the B-spline bases and the rest to the penalty terms along
        each dimension.
    y : np.ndarray
        The response variable sample.

    References
    ----------
    - Golub, G. H., Heath, M., & Wahba, G. (1979). Generalized cross-validation
      as a method for choosing a good ridge parameter. Technometrics, 21(2),
      215-223.

    Returns
    -------
    float
        The GCV value.
    """
    Q = quadratic_term(sp=sp, Q_matrices=Q_matrices)
    # The fitted y
    y_hat = explicit_y_hat(Q=Q, B_weighted=B_weighted, y=y)
    # Return the GCV value, which is RSS * n / (n - tr(H))**2, where RSS is the
    # residual sum of squares, n is the product of the dimensions of y and H is
    # the hat matrix of the unconstrained problem
    return (np.linalg.norm((y - y_hat)) ** 2 * np.prod(y.shape)) / (
        np.prod(y.shape) - np.trace(np.linalg.solve(Q, Q_matrices[0]))
    ) ** 2



import numpy as np
from scipy.linalg import ldl


def cholesky_semidef(A: np.ndarray) -> np.ndarray:

    """Computes a Cholesky decomposition of a positive semidefinite symmetric
    matrix. This is done by applying a LDL decomposition implementation in
    scipy.linalg, which admits any symmetric matrix independent of his
    signature. The lower triangular matrix L is obtained multiplying the L
    matrix in the LDL decomposition by the square root of the diagonal elements
    in D, which are all non-negative due to the positive semidefinitess of the
    input matrix.

    Parameters
    ----------
    A : np.ndarray
        A positive semidefinite symmetric matrix.

    Returns
    -------
    np.ndarray
        The lower triangular matrix of a Cholesky decomposition of the input
        matrix.
    """

    # Compute the LDL decomposition of the matrix
    LDL_decomp = ldl(A, lower=False)
    # Compute the square root of the diagonal of D. In many cases, these values
    # may be negative due to numerical errors, and they are clipped to zero
    sqrt_decomp = np.sqrt(np.clip(np.diag(LDL_decomp[1]), a_min=0, a_max=1e16))
    # Get L in the Cholesky decomposition as sqrt(D) @ L
    return sqrt_decomp * LDL_decomp[0]



from typing import Dict, Iterable, Tuple, Union

import mosek.fusion
import numpy as np
from cpsplines.mosek_functions.utils_mosek import matrix_by_tensor_product_mosek
from cpsplines.psplines.bspline_basis import BsplineBasis


class PointConstraints:

    """
    Define the constraints that the smoothing (or a derivative in a particular
    direction) at a certain point must be bounded around an input value. The
    width of the bound is controlled by a tolerance parameter. All these
    constraints share the derivatives orders enforced.

    Parameters
    ----------
    pts : Iterable[np.ndarray]
        An iterable of arrays containing the coordinates of the points where the
        constraints are enforced. The first vector contain the coordinates in
        the 0-th axis, the second vector the coordinates in the 1-st axis, and
        so on.
    value : mosek.fusion.Model
        The middle point of the bounds that the smoothing (or derivative) must
        fulfill.
    derivative : Iterable[int]
        The orders of the derivatives. The first element corresponds to the
        derivative along the 0-th axis, the second element along the 1-st axis,
        and so on.
    bspline : Iterable[BsplineBasis]
        The B-spline bases objects.
    tolerance : Union[int, float]
        The tolerance used to define the bounds.
    """

    def __init__(
        self,
        pts: Iterable[np.ndarray],
        value: Iterable[Union[int, float]],
        derivative: Iterable[int],
        bspline: Iterable[BsplineBasis],
        tolerance: Union[int, float],
    ):
        self.pts = pts
        self.value = value
        self.derivative = derivative
        self.bspline = bspline
        self.tolerance = tolerance

    def point_cons(
        self,
        var_dict: Dict[str, mosek.fusion.LinearVariable],
        model: mosek.fusion.Model,
    ) -> Tuple[mosek.fusion.LinearConstraint]:

        """
        Constructs the point constraints for the fixed combination of
        derivative orders. These linear constraints are obtained evaluating the
        estimation of the B-splines expansion at the desired points (which yield
        to a linear term depending on elements from the multidimensional
        coefficient array) and setting the difference between this output and
        the value to be less than the tolerance.

        Parameters
        ----------
        var_dict : Dict[str, mosek.fusion.LinearVariable]
            A dictionary containing the decision variables.
        model : mosek.fusion.Model
            The MOSEK model of the problem.

        Returns
        -------
        Tuple[mosek.fusion.LinearConstraint]
            A tuple containing the point constraints.

        Raises
        ------
        ValueError
            If the array of coordinates `pts` and the derivative order vector
            `derivative` have different length than the number of B-spline bases.
        """

        if any(len(x) != len(self.bspline) for x in [self.pts, self.derivative]):
            raise ValueError(
                "`pts` and `derivative` lengths must be equal the number of covariates."
            )
        # Get the evaluations of the coordinates at their respective B-spline
        # basis and the corresponding derivative order
        bsp_eval = {
            i: bsp.bspline_basis.derivative(nu=self.derivative[i])(self.pts[i])
            for i, bsp in enumerate(self.bspline)
        }

        list_cons = []
        # For every point constraint, extract the evaluation of the
        # corresponding coordinates and multiply them by the multidimensional
        # array of the expansion coefficients
        for i, v in enumerate(self.value):
            bsp_x = [np.expand_dims(val[i, :], axis=1).T for val in bsp_eval.values()]
            coef = matrix_by_tensor_product_mosek(
                matrices=bsp_x, mosek_var=var_dict["theta"]
            )
            # The output should be constrained on the interval (v - tol, v + tol)
            list_cons.append(
                model.constraint(
                    coef,
                    mosek.fusion.Domain.greaterThan(v - self.tolerance),
                )
            )
            list_cons.append(
                model.constraint(
                    coef,
                    mosek.fusion.Domain.lessThan(v + self.tolerance),
                )
            )
        return tuple(list_cons)



import pathlib
from setuptools import find_packages, setup

README = (pathlib.Path(__file__).parent / "README.md").read_text()

setup(
    name="cpsplines",
    license="MIT",
    version="0.1.1",
    packages=find_packages(),
    description="Constrained P-splines",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Manuel Navarro García",
    author_email="manuelnavarrogithub@gmail.com",
    url="https://github.com/ManuelNavarroGarcia/cpsplines/",
    # download_url="https://github.com/ManuelNavarroGarcia/cpsplines/archive/refs/tags/0.1.tar.gz",
    keywords=["P-splines", "MOSEK", "Python 3.9", "Constraints", "Optimization"],
    python_requires=">=3.9",
    setup_requires=["setuptools_scm"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=[
        "joblib",
        "matplotlib",
        "mosek",
        "numpy",
        "pandas",
        "scipy",
        "tensorly",
    ],
    extras_require={"dev": ["black", "ipykernel", "mypy", "pip-tools", "pytest"]},
)



import numpy as np
import math
from scipy.interpolate import BSpline
from typing import Dict, List, Union


class BsplineBasis:

    """
    Generate a univariate uniform B-spline basis from a regressor vector. The
    knot sequence is defined using `n_int + 2 * deg + 1` evenly spaced knots so
    that the (`deg` + 1)-th knot coincides with the minimum value of `xsample` and
    the (`n_int` + `deg` + 1) matches the maximum value of `xsample`. The B-spline
    basis may be extended outwards the range defined by `xsample` adding extra
    knots (preserving the same step length) to the knot sequence as it is done
    in Currie I., Durban M. and Eilers P. (2004).

    Parameters
    ----------
    xsample : np.ndarray
        The regressor vector. Must be a unidimensional array.
    deg : int, optional
        The polynomial degree of the B-spline basis. Must be a non-negative
        integer. By default, 3.
    n_int : int, optional
        The number of equal intervals which [min(`xsample`), max(`xsample`)] is
        split. Must be greater or equal than 2. By default, 40.
    prediction : Dict[str, Union[int, float]], optional
        A dictionary containing the most extreme values that the extended
        basis needs to capture. The keys are 'backwards' and 'forward',
        depending on the direction which the basis must be extended, and the
        values are the points on the basis axis. If 'backwards' ('forward') keys
        is not empty, the value must be at the left (right) of the minimum
        (maximum) value of 'xsample'. By default, {}.

    Attributes
    ----------
    int_back : int
        The number of extra knots used to extend the B-spline basis to the left.
    int_forw : int
        The number of extra knots used to extend the B-spline basis to the right.
    knots : np.ndarray of shape (`n_int` + 2 * `deg` + 1, )
        The knot sequence of the B-spline basis.
    bspline_basis : scipy.interpolate.Bspline
        The `n_int` + `int_back` + `int_forw` + `deg` elements of the B-spline basis.
    matrixB : np.ndarray of shape (`int_back` + len(`xsample`) + `int_forw`,
        `n_int` + `deg`)
        The design matrix of the B-spline basis. The ij-th element of the matrix
        contains the evaluation of the j-th B-spline from the basis at the i-th
        element of the vector `xsample`.

    References
    ----------
    - Currie, I. D., Durban, M., & Eilers, P. H. (2004). Smoothing and
      forecasting mortality rates. Statistical modelling, 4(4), 279-298.
    - De Boor, C., & De Boor, C. (1978). A practical guide to splines (Vol. 27).
      New York: Springer-Verlag.
    """

    def __init__(
        self,
        xsample: np.ndarray,
        deg: int = 3,
        n_int: int = 40,
        prediction: Dict[str, Union[int, float]] = {},
    ):
        self.xsample = np.sort(xsample)
        self.deg = deg
        self.n_int = n_int
        self.prediction = prediction

    def get_matrix_B(self) -> None:

        """
        Defines the design matrix of the B-spline basis, consisting on the
        evaluation of the B-spline basis polynomials at `xsample`. This is
        achieved determining how many extra knots are needed to extend the basis
        on the desired range, defining the knot sequence of the basis, which is
        constructed using `construct_fast` from scipy.interpolate.Bspline, and
        finally evaluating this basis at `xsample`. Hence, the shape of the
        matrix is  (`int_back` + len(`xsample`) + `int_forw`, `n_int` + `deg`)

        Raises
        ------
        ValueError
            If `deg` is a negative integer.
        ValueError
            If `xsample` is not a unidimensional array.
        ValueError
            If `n_int` is less than 2.
        ValueError
            If any key in `prediction` is different from 'forward' or
            'backwards'.
        ValueError
            If the bound given in 'backwards' is inside the convex hull of
            `xsample`.
        ValueError
            If the bound given in 'forward' is inside the convex hull of
            `xsample`.
        """

        if self.deg < 0:
            raise ValueError("The degree of the B-spline basis must be at least 0.")
        if self.xsample.ndim != 1:
            raise ValueError("Regressor vector must be one-dimensional.")
        if self.n_int < 2:
            raise ValueError(
                "The fitting regions must be split in at least 2 intervals."
            )
        if len(set(self.prediction) - set(["backwards", "forward"])) > 0:
            raise ValueError(
                "Prediction only admits as keys `forward` and `backwards`."
            )

        if "backwards" in self.prediction:
            if self.prediction["backwards"] >= self.xsample[0]:
                raise ValueError(
                    (
                        "Backwards prediction limit must stand on the "
                        "left-hand side of the regressor vector."
                    )
                )
        if "forward" in self.prediction:
            if self.prediction["forward"] <= self.xsample[-1]:
                raise ValueError(
                    (
                        "Forward prediction limit must stand on the "
                        "right-hand side of the regressor vector."
                    )
                )

        # Compute the distance between adjacent knots
        step_length = (self.xsample[-1] - self.xsample[0]) / self.n_int

        # Determine how many knots at the left (right) of min(`xsample`)
        # (max(`xsample`)) are needed to extend the basis backwards (forward).
        # The step length between these new knots must be the same
        self.int_back = (
            math.ceil((self.xsample[0] - self.prediction["backwards"]) / step_length)
            if "backwards" in self.prediction
            else 0
        )

        self.int_forw = (
            math.ceil((self.prediction["forward"] - self.xsample[-1]) / step_length)
            if "forward" in self.prediction
            else 0
        )
        # Construct the knot sequence of the B-spline basis, consisting on
        # `n_int` + 2 * `deg` + 1 equally spaced knots
        knots = np.linspace(
            self.xsample[0] - (self.int_back + self.deg) * step_length,
            self.xsample[-1] + (self.int_forw + self.deg) * step_length,
            self.n_int + self.int_back + self.int_forw + 2 * self.deg + 1,
        )
        # To avoid floating point error, force that the (`deg` + 1)-th knot
        # coincides with the minimum value of `xsample` and the
        # (`n_int` + `deg` + 1) matches the maximum value of `xsample`
        knots[self.int_back + self.deg] = self.xsample[0]
        knots[-(self.int_forw + self.deg + 1)] = self.xsample[-1]
        self.knots = knots

        # Construct the B-spline basis, consisting on
        # (`n_int` + `int_back` + `int_forw` + `deg`) elements
        self.bspline_basis = BSpline.construct_fast(
            t=self.knots,
            c=np.eye(
                self.n_int + self.deg + self.int_forw + self.int_back, dtype=float
            ),
            k=self.deg,
        )
        # Return the design matrix of the B-spline basis
        self.matrixB = self.bspline_basis.derivative(nu=0)(
            np.concatenate(
                [
                    self.knots[self.deg : self.deg + self.int_back],
                    self.xsample,
                    self.knots[self.int_back + self.n_int + self.deg + 1 : -self.deg],
                ]
            )
        )
        return None

    def get_matrices_S(self) -> List[np.ndarray]:

        """
        Generate a list of matrices (one by interval) containing the polynomial
        coefficients from the non-zero B-spline basis elements . For the j-th
        matrix, the first row contains the independent coefficients from the
        (j-`deg`)-th to the j-th B-spline basis elements, the second row
        contains the linear coefficients from the polynomials, and so on.

        These matrices are required to enforce non-negativity constraints using
        Proposition 1 in Bertsimas and Popescu (2002).

        References
        ----------
        - Bertsimas, D., & Popescu, I. (2002). On the relation between option
          and stock prices: a convex optimization approach. Operations Research,
          50(2), 358-374.

        Returns
        -------
        List[np.ndarray]
            List of matrices of shape (`deg` + 1, `deg` + 1) with the polynomial
            coefficients.
        """

        # Since knots are evenly spaced, the value of the B-spline basis
        # elements is periodic with period the step length of the knot sequence.
        # Hence, we compute the value of the non-zero basis elements at only one
        # interval (take advantage of the periodicity) at `deg` + 1 points.
        # These points should be located at the same interval, and we choose
        #          linspace(knots[`deg`], knots[`deg` + 1], `deg` + 1),
        # since this is the first interval with (`deg` + 1) non-zero elements.
        C = np.zeros(shape=(self.deg + 1, self.deg + 1))
        for i in range(self.deg + 1):
            C[:, i] = BSpline.basis_element(t=self.knots[i : self.deg + 2 + i])(
                x=np.linspace(
                    self.knots[self.deg], self.knots[self.deg + 1], self.deg + 1
                )
            )
        S = []
        # The matrices S_k that we are looking for satisfy that S_k @ T_k = C,
        # where T_k has as columns the array (1, x, x**2, ... x**`deg`)
        # evaluated at the points
        #     linspace(knots[k + `deg`], knots[k + `deg` + 1], `deg` + 1)
        for k in range(self.n_int + self.int_back + self.int_forw):
            T_k = np.vander(
                np.linspace(
                    self.knots[k + self.deg], self.knots[k + self.deg + 1], self.deg + 1
                ),
                increasing=True,
            )
            S_k = np.linalg.solve(T_k, C)
            S.append(S_k)
        return S



import itertools
from typing import Dict, Iterable, List, Tuple, Union

import mosek.fusion
import numpy as np
from cpsplines.mosek_functions.utils_mosek import matrix_by_tensor_product_mosek
from cpsplines.psplines.bspline_basis import BsplineBasis
from scipy.special import comb, factorial


class IntConstraints:

    """
    Define the set of constraints that ensures the non-negativity (or
    non-positivity) for the derivative order `derivative`along the respective
    axis of the variable `var_name`. For the unidimensional case, the
    requirements are completely fulfilled for every point in the range where the
    B-spline basis is defined. For the multivariate case, the constraints are
    imposed at the curves that pass through the inner knots of the B-spline
    basis. For all cases, the constraints are imposed following the Proposition
    1 in Bertsimas and Popescu (2002).

    Parameters
    ----------
    bspline : Iterable[BsplineBasis]
        An iterable containing the B-spline bases objects used to approximate
        the function to estimate.
    var_name : int
        The name of the variable along the constraints are imposed. Must be a
        non-negative integer.
    derivative : int
        The derivative order of the function that needs to be constrained. Must
        be a non-negative integer.
    constraints : Dict[str, Union[int, float]]
        The constraints that are imposed. The keys of the dictionary are either
        "+" (the function (or some derivative) needs to be above some threshold)
        or "-" (the function (or some derivative) needs to be below some
        threshold), and the values are the thresholds.

    Attributes
    ----------
    deg_w : int
        The shape of the matrices W.
    matricesW : List[np.ndarray] of shape (deg_w, deg_w)
        The matrices W used to impose the constraints by virtue of Proposition 1
        in Bertsimas and Popescu (2002).
    matricesH : mosek.fusion.Matrix
        Auxiliary matrices to extract the correct coefficients from the positive
        semidefinite matrix variable in the left-hand side from the equations of
        Proposition 1 in Bertsimas and Popescu (2002).

    References
    ----------
    - Bertsimas, D., & Popescu, I. (2002). On the relation between option
      and stock prices: a convex optimization approach. Operations Research,
      50(2), 358-374.
    """

    def __init__(
        self,
        bspline: Iterable[BsplineBasis],
        var_name: int,
        derivative: int,
        constraints: Dict[str, Union[int, float]],
    ):
        self.bspline = bspline
        self.var_name = var_name
        self.derivative = derivative
        self.constraints = constraints

    def _get_matrices_W(self) -> List[np.ndarray]:

        """
        Generates matrices W containing the weights that have to be multiplied
        by the coefficients of the polynomials from the B-spline basis to employ
        Proposition 1 in Bertsimas and Popescu (2002). For each interval of the
        B-spline basis, one matrix W is defined.

        Returns
        -------
        List[np.ndarray]
            The list of matrices W.

        Raises
        ------
        ValueError
            If `var_name` is not a non-negative integer.
        ValueError
            If `derivative` is not a non-negative integer.
        ValueError
            If `deg_w` is not a non-negative integer.
        """

        if self.var_name < 0:
            raise ValueError("The variable name must be a non-negative integer.")
        if self.derivative < 0:
            raise ValueError("The derivative order must be a non-negative integer.")

        # Select the B-spline basis along the corresponding axis
        bsp = self.bspline[self.var_name]
        # Define deg_w, which coincides with the order of the matrices W
        if bsp.deg - self.derivative < 0:
            raise ValueError(
                "The derivative order must be lower than the B-spline basis degree."
            )
        self.deg_w = bsp.deg - self.derivative
        # Deriving a polynomial gives rise to a coefficient vector whose length
        # reduces by one the degree of the polynomial (only non-zero
        # coefficients are considered). Recursively, a polynomial with degree d
        # once derivated n times will leave a d-n vector. This vector is given
        # by the n column and the first (n-d) rows of the symmetric Pascal
        # matrix of order d.
        pascal_coef = np.array(
            [
                factorial(self.derivative) * comb(self.derivative + i, self.derivative)
                for i in range(self.deg_w + 1)
            ],
            dtype=np.int32,
        )
        W = []
        # Generate the matrices W, containing the weights from the Proposition 1
        # of Bertsimas and Popescu (2002)
        for i in range(bsp.matrixB.shape[1] - bsp.deg):
            W_i = np.zeros(shape=(self.deg_w + 1, self.deg_w + 1))
            for q in range(self.deg_w + 1):
                for m in range(q + 1):
                    for r in range(m, self.deg_w + 1 + m - q):
                        W_i[q, r] += (
                            comb(r, m)
                            * comb(self.deg_w - r, q - m)
                            * (bsp.knots[bsp.deg : -bsp.deg][i]) ** (r - m)
                            * (bsp.knots[bsp.deg : -bsp.deg][i + 1]) ** (m)
                        )
            # We multiply this vector coefficient by each matrix W instead of
            # multiplying it by matrices S, which contain the non-zero
            # coefficients of the polynomials, since at the end of the day we
            # will take the product WS on the axis where constraints are
            # enforced
            W.append(pascal_coef * W_i)
        return W

    def _get_matrices_H(self) -> List[List[mosek.fusion.Matrix]]:

        """
        Generates matrices H used to extract the right coefficients from the
        positive semidefinite matrix variables to fulfill the equations of
        Proposition 1 in Bertsimas and Popescu (2002). This is done by using the
        Frobenius inner product and, as it is stated in the proposition,
        matrices have order `deg_w`+ 1. For each interval of the B-spline basis,
        one matrix H is defined.

        Returns
        -------
        List[List[mosek.fusion.Matrix]]
            A list containing two list of matrices: the first one corresponds to
            the matrices used on the homogeneous equations, while the second is
            used on the non-homogeneous equations. For the last, the first
            matrix corresponds to the independent term equation, the second to
            the linear term equation, and so on.
        """

        # The elements from X in the homogeneous equations are located on its
        # even antidiagonals, while the ones from the non-homogeneous equations
        # are situated on the odd antidiagonals
        diag_zero = np.linspace(
            1 - self.deg_w, self.deg_w - 1, self.deg_w, dtype=np.int8
        )
        diag_nonzero = np.linspace(
            -self.deg_w, self.deg_w, self.deg_w + 1, dtype=np.int8
        )
        H = []
        for diag in [diag_zero, diag_nonzero]:
            H_by_diag = []
            for k in diag:
                # Create an identity matrix along the corresponding diagonal and
                # the rotate it 90 degrees to get the antidiagonal. Then convert
                # it to a MOSEK sparse matrix
                H_by_diag.append(
                    mosek.fusion.Matrix.sparse(
                        np.rot90(np.eye(self.deg_w + 1, k=k, dtype=np.int32))
                    )
                )
            H.append(H_by_diag)
        return H

    def _create_PSD_var(self, model: mosek.fusion.Model) -> mosek.fusion.PSDVariable:

        """
        Create the positive semidefinite matrix variables used to formulate the
        constraints. They have shape (`deg_w` + 1, `deg_w` + 1) and one by
        interval where constraints are applied is needed.

        Parameters
        ----------
        model : mosek.fusion.Model
            The MOSEK model of the problem.

        Returns
        -------
        mosek.fusion.PSDVariable
            A three-dimensional variable with shape (`deg_w` + 1, `deg_w` + 1,
            n * m), where n is the number of interval where the constraint need
            to be imposed, m is the number of sign constraints related to this
            derivative order and variable, and each of the matrices (`deg_w` +
            1, `deg_w` + 1) are required to be positive semidefinite.
        """

        # The number of intervals where the constraint need to be enforced can
        # be computed as the product of interior intervals
        n = [bsp.matrixB.shape[1] - bsp.deg for bsp in self.bspline]
        return model.variable(
            mosek.fusion.Domain.inPSDCone(
                self.deg_w + 1, len(self.constraints.keys()) * np.prod(n)
            )
        )

    def interval_cons(
        self,
        var_dict: Dict[str, mosek.fusion.LinearVariable],
        model: mosek.fusion.Model,
        matrices_S: Dict[int, Iterable[np.ndarray]],
    ) -> Tuple[mosek.fusion.LinearConstraint]:

        """
        Defines the non-negative related constraints over a finite interval. For
        each interval and each sign constraint, 2 * `deg_w` - 1 equations are
        defined.

        Parameters
        ----------
        var_dict :  Dict[str, mosek.fusion.LinearVariable]
            The dictionary that contains the decision variables used to define
            the objective function of the problem.
        model : mosek.fusion.Model
            The MOSEK model of the problem.
        matrices_S : Dict[int, Iterable[np.ndarray]]
            Contains as keys the indices of the variable, and as values the
            matrices S.

        Returns
        -------
        Tuple[mosek.fusion.LinearConstraint]
            The set of constraints along the variable `var_name` and derivative
            order `derivative`.

        Raises
        ------
        ValueError
            If the constraints are not either "+" or "-".
        """

        if len(set(self.constraints.keys()) - set(["+", "-"])) > 0:
            raise ValueError("The constraint sign must be either `+` or `-`.")

        # Generate the matrices H and W, and the decision variables X
        self.matricesW = self._get_matrices_W()
        self.matricesH = self._get_matrices_H()
        var_dict |= {"X": self._create_PSD_var(model=model)}
        # Extract the weights related to the independent term, which is used
        # when some derivative of the curve is constrained to be greater or
        # lower than a certain threshold
        ind_term = self.matricesW[0][:, 0]
        # For every axis, get the contribution to the estimated function at the
        # knots
        for j, bsp in enumerate(self.bspline):
            # The contribution along the direction where the constraints are
            # imposed is W (that already include the vector coefficient arised
            # when differentiate the polynomial) times S once the first
            # `derivative` rows are deleted
            if self.var_name == j:
                matrices_S[self.var_name] = [
                    self.matricesW[i] @ np.delete(s, range(self.derivative), axis=0)
                    for i, s in enumerate(matrices_S[self.var_name])
                ]
            # Since the knot sequence is evenly spaced, the value of the
            # B-splines is periodic an it is always the same, so we pick up the
            # value of all the B-splines at the first knot
            else:
                value_at_knots = np.expand_dims(
                    bsp.matrixB[
                        bsp.int_back, bsp.int_back : bsp.int_back + bsp.deg + 1
                    ],
                    axis=0,
                )
                matrices_S[j] = [value_at_knots for _ in range(len(matrices_S[j]))]
        # For every interval on the `var_name` axis, count how many interval
        # constraints of the same sign and fixed derivative need to be
        # considered
        num_by_interval = (
            np.prod(
                np.array(
                    [
                        bsp.matrixB.shape[1] - bsp.deg
                        for i, bsp in enumerate(self.bspline)
                        if i != self.var_name
                    ]
                )
            )
            if len(matrices_S.keys()) > 1
            else 1
        )
        # For every interval on the `var_name` axis, count how many interval
        # constraints types by sign there are
        num_by_sign = len(self.constraints.keys())
        list_cons = []
        # Loop over every interval on the `var_name` axis
        for w in range(
            self.bspline[self.var_name].matrixB.shape[1]
            - self.bspline[self.var_name].deg
        ):
            # Create a list containing the lists with the contribution to the
            # estimated function at the knots
            a = [v for i, v in matrices_S.items() if i != self.var_name]
            # Create a list containing ranges of the same length as previous list
            a_idx = [range(len(v)) for i, v in matrices_S.items() if i != self.var_name]
            # Insert at the position `var_name` the correct value/index of S
            # along the `var_name` direction
            a.insert(self.var_name, [matrices_S[self.var_name][w]])
            a_idx.insert(self.var_name, range(w, w + 1))
            # Generate all the combinations possible from previous lists
            iter_a = list(itertools.product(*a))
            iter_idx = list(itertools.product(*a_idx))
            # Loop over the size of all the combinations (same length as
            # positive semidefinite variables) and their corresponding values of
            # the contribution
            for j, (id, mat) in enumerate(zip(iter_idx, iter_a)):
                last_id = [id[i] + bsp.deg + 1 for i, bsp in enumerate(self.bspline)]
                # Slice the multidimensional coefficient variable on the
                # interval on the values the corresponding B-spline is non-zero
                coef_theta = var_dict["theta"].slice(
                    np.array(id, dtype=np.int32), np.array(last_id, dtype=np.int32)
                )
                # Multiply the sliced variable on each face by the correct
                # contribution
                poly_coef = mosek.fusion.Expr.flatten(
                    matrix_by_tensor_product_mosek(matrices=mat, mosek_var=coef_theta)
                )
                # Loop over the different sign constraints
                for k, key in enumerate(self.constraints.keys()):
                    # Get current index
                    actual_index = k + num_by_sign * (j + w * num_by_interval)
                    # Get current positive semidefinite variable
                    slice_X = (
                        var_dict["X"]
                        .slice(
                            [actual_index, 0, 0],
                            [
                                actual_index + 1,
                                self.deg_w + 1,
                                self.deg_w + 1,
                            ],
                        )
                        .reshape([self.deg_w + 1, self.deg_w + 1])
                    )
                    # Creates the homogeneous equations
                    for i in range(self.deg_w):
                        list_cons.append(
                            model.constraint(
                                mosek.fusion.Expr.dot(self.matricesH[0][i], slice_X),
                                mosek.fusion.Domain.equalsTo(0.0),
                            )
                        )
                    # Creates the non-homogeneous equations
                    for i in range(self.deg_w + 1):
                        if key == "+":
                            list_cons.append(
                                model.constraint(
                                    mosek.fusion.Expr.sub(
                                        poly_coef.slice(i, i + 1),
                                        mosek.fusion.Expr.dot(
                                            self.matricesH[1][i], slice_X
                                        ),
                                    ),
                                    mosek.fusion.Domain.equalsTo(
                                        ind_term[i] * self.constraints[key]
                                    ),
                                )
                            )
                        else:
                            list_cons.append(
                                model.constraint(
                                    mosek.fusion.Expr.add(
                                        poly_coef.slice(i, i + 1),
                                        mosek.fusion.Expr.dot(
                                            self.matricesH[1][i], slice_X
                                        ),
                                    ),
                                    mosek.fusion.Domain.equalsTo(
                                        ind_term[i] * self.constraints[key]
                                    ),
                                )
                            )
        return tuple(list_cons)
